/**
 * Copyright (c) UT-Battelle, LLC. 2014-2017. ALL RIGHTS RESERVED.
 * See file LICENSE for terms.
 */

#ifndef UCT_UGNI_CONTEXT_H
#define UCT_UGNI_CONTEXT_H

#include "ugni_types.h"
#include "ugni_def.h"
#include <uct/base/uct_md.h>

extern uct_component_t uct_ugni_component;

#endif
