/**
* Copyright (C) Mellanox Technologies Ltd. 2001-2014.  ALL RIGHTS RESERVED.
*
* See file LICENSE for terms.
*/

#ifndef RC_DEF_H_
#define RC_DEF_H_


typedef struct uct_rc_ep                uct_rc_ep_t;
typedef struct uct_rc_iface             uct_rc_iface_t;
typedef struct uct_rc_iface_send_op     uct_rc_iface_send_op_t;
typedef struct uct_rc_iface_send_desc   uct_rc_iface_send_desc_t;
typedef struct uct_rc_iface_config      uct_rc_iface_config_t;
typedef struct uct_rc_pending_priv      uct_rc_pending_priv_t;
typedef struct uct_rc_txqp              uct_rc_txqp_t;


#endif
